from abc import abstractmethod
import numpy as np
import tensorflow as tf
from tensorflow.python.ops import random_ops
from tensorflow.python.ops import stateless_random_ops
from tensorflow.keras.initializers import Initializer
from cvnn import logger
import sys
from pdb import set_trace
# Typing
from typing import Optional

INIT_TECHNIQUES = {'zero_imag', 'mirror'}


def _compute_fans(shape):
    """
    Taken from https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/ops/init_ops_v2.py#L994
    Computes the number of input and output units for a weight shape.
    Args:
        shape: Integer shape tuple or TF tensor shape.
    Returns:
        A tuple of scalars (fan_in, fan_out).
    """
    if len(shape) < 1:  # Just to avoid errors for constants.
        fan_in = fan_out = 1.
    elif len(shape) == 1:
        fan_in = fan_out = shape[0]
    elif len(shape) == 2:
        fan_in = shape[0]
        fan_out = shape[1]
    else:
        # Assuming convolution kernels (2D, 3D, or more).
        # kernel shape: (..., input_depth, depth)
        receptive_field_size = 1.
        for dim in shape[:-2]:
            receptive_field_size *= dim
        fan_in = shape[-2] * receptive_field_size
        fan_out = shape[-1] * receptive_field_size
    return fan_in, fan_out


class _RandomGenerator(object):
    """
    Random generator that selects appropriate random ops.
    https://github.com/tensorflow/tensorflow/blob/2b96f3662bd776e277f86997659e61046b56c315/tensorflow/python/ops/init_ops_v2.py#L1041
    """

    def __init__(self, seed=None):
        super(_RandomGenerator, self).__init__()
        if seed is not None:
            # Stateless random ops requires 2-int seed.
            self.seed = [seed, 0]
        else:
            self.seed = None

    def random_normal(self, shape, mean=0.0, stddev=1, dtype=tf.dtypes.float32):
        """A deterministic random normal if seed is passed."""
        if self.seed:
            op = stateless_random_ops.stateless_random_normal
        else:
            op = random_ops.random_normal
        return op(shape=shape, mean=mean, stddev=stddev, dtype=dtype, seed=self.seed)

    def random_uniform(self, shape, minval, maxval, dtype):
        """A deterministic random uniform if seed is passed."""
        if self.seed:
            op = stateless_random_ops.stateless_random_uniform
        else:
            op = random_ops.random_uniform
        return op(shape=shape, minval=minval, maxval=maxval, dtype=dtype, seed=self.seed)

    def truncated_normal(self, shape, mean, stddev, dtype):
        """A deterministic truncated normal if seed is passed."""
        if self.seed:
            op = stateless_random_ops.stateless_truncated_normal
        else:
            op = random_ops.truncated_normal
        return op(shape=shape, mean=mean, stddev=stddev, dtype=dtype, seed=self.seed)


class ComplexInitializer(Initializer):

    def __init__(self, distribution: str = "uniform", seed: Optional[int] = None):
        if distribution.lower() not in {"uniform", "normal"}:
            raise ValueError("Invalid `distribution` argument:", distribution)
        else:
            self.distribution = distribution.lower()
        self._random_generator = _RandomGenerator(seed)

    def _call_random_generator(self, shape, arg, dtype):
        if self.distribution == "uniform":
            return self._random_generator.random_uniform(shape=shape, minval=-arg, maxval=arg, dtype=dtype)
        elif self.distribution == "normal":
            # I make this magic number division because that's what tf does on this case
            return self._random_generator.truncated_normal(shape=shape, mean=0.0, stddev=arg / .87962566103423978,
                                                           dtype=dtype)

    @abstractmethod
    def _compute_limit(self, fan_in, fan_out):
        pass

    def __call__(self, shape, dtype=tf.dtypes.complex64, **kwargs):
        fan_in, fan_out = _compute_fans(shape)
        arg = self._compute_limit(fan_in, fan_out)
        dtype = tf.dtypes.as_dtype(dtype)
        if dtype.is_complex:
            arg = arg / np.sqrt(2)
        return self._call_random_generator(shape=shape, arg=arg, dtype=dtype.real_dtype)

    def get_config(self):  # To support serialization
        return {"seed": self._random_generator.seed}


class ComplexGlorotUniform(ComplexInitializer):
    """
    The Glorot uniform initializer, also called Xavier uniform initializer.
    Reference: http://proceedings.mlr.press/v9/glorot10a.html
    Draws samples from a uniform distribution:
        - Real case: `x ~ U[-limit, limit]` where `limit = sqrt(6 / (fan_in + fan_out))`
        - Complex case: `z / Re{z} = Im{z} ~ U[-limit, limit]` where `limit = sqrt(3 / (fan_in + fan_out))`
    where `fan_in` is the number of input units in the weight tensor and `fan_out` is the number of output units.

    ```
    # Standalone usage:
    import cvnn
    initializer = cvnn.initializers.ComplexGlorotUniform()
    values = initializer(shape=(2, 2))                  # Returns a complex Glorot Uniform tensor of shape (2, 2)
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.ComplexGlorotUniform()
    layer = cvnn.layers.ComplexDense(units=10, kernel_initializer=initializer)
    ```
    """
    __name__ = "Complex Glorot Uniform"

    def __init__(self, seed: Optional[int] = None):
        super(ComplexGlorotUniform, self).__init__(distribution="uniform", seed=seed)

    def _compute_limit(self, fan_in, fan_out):
        return tf.math.sqrt(6. / (fan_in + fan_out))


class ComplexGlorotNormal(ComplexInitializer):
    """
    The Glorot normal initializer, also called Xavier normal initializer.
    Reference: http://proceedings.mlr.press/v9/glorot10a.html
        *Note: The reference actually refers to the uniform case but it's analysis was adapted for a normal distribution
    Draws samples from a truncated normal distribution centered on 0 with
     - Real case: `stddev = sqrt(2 / (fan_in + fan_out))`
     - Complex case: real part stddev = complex part stddev = `1 / sqrt(fan_in + fan_out)`
    where `fan_in` is the number of input units in the weight tensor and `fan_out` is the number of output units.

    ```
    # Standalone usage:
    import cvnn
    initializer = cvnn.initializers.ComplexGlorotNormal()
    values = initializer(shape=(2, 2))                  # Returns a complex Glorot Normal tensor of shape (2, 2)
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.ComplexGlorotNormal()
    layer = cvnn.layers.ComplexDense(units=10, kernel_initializer=initializer)
    ```
    """
    __name__ = "Complex Glorot Normal"

    def __init__(self, seed: Optional[int] = None):
        super(ComplexGlorotNormal, self).__init__(distribution="normal", seed=seed)

    def _compute_limit(self, fan_in, fan_out):
        return tf.math.sqrt(2. / (fan_in + fan_out))


class ComplexHeUniform(ComplexInitializer):
    """
    The He Uniform initializer.
    Reference: http://proceedings.mlr.press/v9/glorot10a.html
    Draws samples from a uniform distribution:
        - Real case: `x ~ U[-limit, limit]` where `limit = sqrt(6 / fan_in)`
        - Complex case: `z / Re{z} = Im{z} ~ U[-limit, limit]` where `limit = sqrt(3 / fan_in)`
    where `fan_in` is the number of input units in the weight tensor.

    ```
    # Standalone usage:
    import cvnn
    initializer = cvnn.initializers.ComplexHeUniform()
    values = initializer(shape=(2, 2))                  # Returns a real He Uniform tensor of shape (2, 2)
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.ComplexHeUniform()
    layer = cvnn.layers.ComplexDense(units=10, kernel_initializer=initializer)
    ```
    """
    __name__ = "Complex He Uniform"

    def __init__(self, seed: Optional[int] = None):
        super(ComplexHeUniform, self).__init__(distribution="uniform", seed=seed)

    def _compute_limit(self, fan_in, fan_out):
        return tf.math.sqrt(6. / fan_in)


class ComplexHeNormal(ComplexInitializer):
    """
    He normal initializer.
    Reference: https://www.cv-foundation.org/openaccess/content_iccv_2015/html/He_Delving_Deep_into_ICCV_2015_paper.html
    It draws samples from a truncated normal distribution centered on 0 with
        - Real case: `stddev = sqrt(2 / fan_in)`
        - Complex case: real part stddev = complex part stddev = `1 / sqrt(fan_in)`
    where fan_in is the number of input units in the weight tensor.

    ```
    # Standalone usage:
    import cvnn
    initializer = cvnn.initializers.ComplexHeNormal()
    values = initializer(shape=(2, 2))                  # Returns a complex He Normal tensor of shape (2, 2)
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.ComplexHeNormal()
    layer = cvnn.layers.ComplexDense(units=10, kernel_initializer=initializer)
    ```
    """
    __name__ = "Complex He Normal"

    def __init__(self, seed: Optional[int] = None):
        super(ComplexHeNormal, self).__init__(distribution="normal", seed=seed)

    def _compute_limit(self, fan_in, fan_out):
        return tf.math.sqrt(2. / fan_in)


class Zeros:
    """
    Creates a tensor with all elements set to zero.

    ```
    > >> cvnn.initializers.Zeros()(shape=(2,2))
    <tf.Tensor: shape=(2, 2), dtype=float32, numpy=
    array([[0.+0.j, 0.+0.j],
          [0.+0.j, 0.+0.j]], dtype=float32)>
    ```

    ```
    # Usage in a cvnn layer:
    import cvnn
    initializer = cvnn.initializers.Zeros()
    layer = cvnn.layers.ComplexDense(units=10, bias_initializer=initializer)
    ```
    """
    __name__ = "Zeros"

    def __call__(self, shape, dtype=tf.dtypes.complex64):
        return tf.zeros(shape, dtype=tf.dtypes.as_dtype(dtype).real_dtype)


class Ones:
    __name__ = "Ones"

    def __call__(self, shape, dtype=tf.dtypes.complex64):
        return tf.ones(shape, dtype=tf.dtypes.as_dtype(dtype).real_dtype)


if __name__ == '__main__':
    # Nothing yet
    set_trace()

__author__ = 'J. Agustin BARRACHINA'
__version__ = '0.0.12'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'



import tensorflow as tf
from tensorflow.keras.layers import Layer
from typing import Optional, Union, Tuple
from cvnn.layers.core import ComplexLayer
from cvnn.layers.core import DEFAULT_COMPLEX_TYPE


class ComplexUpSampling2D(Layer, ComplexLayer):

    def __init__(self, size: Union[int, Tuple[int, int]] = (2, 2),
                 data_format: Optional[str] = None, interpolation: str = 'nearest',
                 align_corners: bool = False, dtype=DEFAULT_COMPLEX_TYPE, **kwargs):
        """
        :param size: Int, or tuple of 2 integers. The upsampling factors for rows and columns.
        :param data_format: string, one of channels_last (default) or channels_first.
            The ordering of the dimensions in the inputs. channels_last corresponds to inputs with shape
            (batch_size, height, width, channels) while channels_first corresponds to inputs with shape
            (batch_size, channels, height, width).
        :param interpolation: A string, one of nearest or bilinear.
        :param align_corners:  if True, the corner pixels of the input and output tensors are aligned,
            and thus preserving the values at those pixels.
            Example of align coreners: https://discuss.pytorch.org/t/what-we-should-use-align-corners-false/22663/9
        """
        self.my_dtype = tf.dtypes.as_dtype(dtype)
        super(ComplexUpSampling2D, self).__init__(dtype=self.my_dtype.real_dtype, **kwargs)
        self.align_corners = align_corners
        if isinstance(size, int):
            self.factor_upsample = (size,) * 2
        else:
            self.factor_upsample = tuple(size)  # Python will tell me if this is not possible
        # TODO: Check is tuple of ints and no negative values!
        self.interpolation = interpolation.lower()
        if self.interpolation not in {'nearest', 'bilinear'}:
            raise ValueError('`interpolation` argument should be one of `"nearest"` or `"bilinear"`.')
        if data_format is None:
            data_format = 'channels_last'
        self.data_format = data_format.lower()
        if self.data_format not in {'channels_first', 'channels_last'}:
            raise ValueError(f'The `data_format` argument must be one of "channels_first", "channels_last". '
                             f'Received: {self.data_format}')

    def call(self, inputs, **kwargs):
        if self.data_format == 'channels_last':
            inputs = tf.transpose(inputs, perm=[1, 2, 0, 3])
        elif self.data_format == 'channels_first':  # I checked it at init, shall I check again?
            inputs = tf.transpose(inputs, perm=[2, 3, 0, 1])
        else:
            raise ValueError(f'The `data_format` argument must be one of "channels_first", "channels_last". '
                             f'Received: {self.data_format}')
        output = self.upsample(inputs=inputs)
        if self.data_format == 'channels_last':
            output = tf.transpose(output, perm=[2, 0, 1, 3])
        elif self.data_format == 'channels_first':  # I checked it at init, shall I check again?
            output = tf.transpose(output, perm=[2, 3, 0, 1])
        else:
            raise ValueError(f'The `data_format` argument must be one of "channels_first", "channels_last". '
                             f'Received: {self.data_format}')
        return output

    def upsample(self, inputs):
        if inputs.dtype.is_integer:  # TODO: Check input is a tensor?
            inputs = tf.cast(inputs, dtype=tf.float32)
        desired_size = [i * o for i, o in zip(inputs.shape, self.factor_upsample)]
        assert len(desired_size) == 2  # The for will do only for the shortest so I should be Ok.
        i_output = tf.reshape(tf.constant([], dtype=inputs.dtype),
                              (0, desired_size[1], tf.shape(inputs)[2], tf.shape(inputs)[3]))
        j_output = tf.reshape(tf.constant([], dtype=inputs.dtype), (1, 0, tf.shape(inputs)[2], tf.shape(inputs)[3]))
        for x in range(0, desired_size[0]):
            for y in range(0, desired_size[1]):
                if self.interpolation == 'bilinear':
                    to_append = self.bilinear(inputs=inputs, x=x, y=y)
                elif self.interpolation == 'nearest':
                    to_append = self.nearest_neighbor(inputs=inputs, x=x, y=y)
                else:
                    raise ValueError(f"Unknown interpolation method {self.interpolation}")
                to_append = tf.expand_dims(tf.expand_dims(to_append, axis=0), axis=0)
                j_output = tf.concat([j_output, to_append], axis=1)
            i_output = tf.concat([i_output, j_output], axis=0)
            j_output = tf.reshape(tf.constant([], dtype=inputs.dtype), (1, 0, tf.shape(inputs)[2], tf.shape(inputs)[3]))
        return i_output

    def nearest_neighbor(self, inputs, x, y):
        # output = tf.repeat(input=tf.repeat(input=inputs, repeats=(self.factor_upsample[0],)*inputs.shape[0],
        #                                    axis=0),
        #                    repeats=(self.factor_upsample[1],)*inputs.shape[1], axis=1)
        # i_new = tf.cast(tf.floor((inputs.shape[0] * x) / desired_output_shape[0]), dtype=tf.int32)
        # j_new = tf.cast(tf.floor((inputs.shape[1] * y) / desired_output_shape[1]), dtype=tf.int32)
        i_new, j_new = self._get_nearest_neighbour(x, y, inputs)
        to_append = inputs[i_new, j_new]
        # assert i_output.shape == (tuple(deisred_size) + (input.shape[2], input.shape[3]))
        return to_append

    def _get_nearest_neighbour(self, x, y, inputs):
        X_small = tf.linspace(0, tf.shape(inputs)[0] - 1, tf.shape(inputs)[0])
        Y_small = tf.linspace(0, tf.shape(inputs)[1] - 1, tf.shape(inputs)[1])
        X, Y = self._to_big(X_small, Y_small)
        i, j = self._get_closest_point(x, y, X, Y)
        return i, j

    @staticmethod
    def _get_closest_point(x, y, x_list, y_list):
        x_distance = tf.math.square(x_list - x)
        y_distance = tf.math.square(y_list - y)
        x_closest = tf.argmin(x_distance)
        y_closest = tf.argmin(y_distance)
        return x_closest, y_closest

    def bilinear(self, inputs, x, y):
        # Equations
        #   https://www.ajdesigner.com/phpinterpolation/linear_interpolation_equation.php
        # Difference with align corners image:
        #   https://discuss.pytorch.org/t/what-we-should-use-align-corners-false/22663/9
        # Examples:
        #   https://www.omnicalculator.com/math/bilinear-interpolation
        #   https://blogs.sas.com/content/iml/2020/05/18/what-is-bilinear-interpolation.html#:~:text=Bilinear%20interpolation%20is%20a%20weighted,the%20point%20and%20the%20corners.&text=The%20only%20important%20formula%20is,x%20%5B0%2C1%5D.
        # Implementations
        #   https://stackoverflow.com/questions/8661537/how-to-perform-bilinear-interpolation-in-python
        (q11, q21, q12, q22), (x1, x2), (y1, y2) = self._get_q_points(x, y, inputs)
        # There are 3 cases:
        #   1. All 4 q's are different and surround the point
        #   2. There are basically 2 q's (get repeated)
        #   3. All 4 q's are equal
        x2_diff = tf.cast(x2, dtype=inputs.dtype) - tf.cast(x, dtype=inputs.dtype)
        x1_diff = tf.cast(x, dtype=inputs.dtype) - tf.cast(x1, dtype=inputs.dtype)
        y2_diff = tf.cast(y2, dtype=inputs.dtype) - tf.cast(y, dtype=inputs.dtype)
        y1_diff = tf.cast(y, dtype=inputs.dtype) - tf.cast(y1, dtype=inputs.dtype)
        delta_x = tf.cast(x2 - x1, dtype=inputs.dtype)
        delta_y = tf.cast(y2 - y1, dtype=inputs.dtype)
        # The next conditions happens in cases 2 or 3, only one for case 2 and both for case 3.
        # Using the following equations/conditions, the general equation stands for all 3 cases.
        if x1 == x2:  # The index was exact, so just make both 1/2
            x2_diff = tf.cast(0.5, dtype=inputs.dtype)
            x1_diff = tf.cast(0.5, dtype=inputs.dtype)
            delta_x = tf.cast(1, dtype=inputs.dtype)
        if y1 == y2:
            y2_diff = tf.cast(0.5, dtype=inputs.dtype)
            y1_diff = tf.cast(0.5, dtype=inputs.dtype)
            delta_y = tf.cast(1, dtype=inputs.dtype)
        t11 = q11 * y2_diff * x2_diff
        t21 = q21 * x1_diff * y2_diff
        t12 = q12 * x2_diff * y1_diff
        t22 = q22 * x1_diff * y1_diff
        to_append = (t11 + t22 + t12 + t21) / (delta_y * delta_x)
        return to_append

    def _get_q_points(self, x, y, inputs):
        # 1. Get x and y coordinates of inputs (basically from 0 to the end)
        X_small = tf.linspace(0, tf.shape(inputs)[0] - 1, tf.shape(inputs)[0])
        Y_small = tf.linspace(0, tf.shape(inputs)[1] - 1, tf.shape(inputs)[1])
        # 2. Transform those coordinates into a index equivalent on the new bigger image.
        X, Y = self._to_big(X_small, Y_small)
        # 3. Get the closest points of X to x. x2 is the closest but bigger and x1 is closest but small
        # for example X = [0.5, 2.5 4.5] then
        #   If x is 1, then x1 is 0.5 and x2 is 2.5.
        #   If x is 2.5, both x1 and x2 are 2.5.
        #   If x is 0, then x1 equals x2.
        #   If x is 5 then x2 equals x1
        x2, x1, y2, y1 = self._get_4_closest_points(x, y, X, Y)
        # Get the points according to the coordinates obtained.
        q11 = inputs[tf.where(x1 == X)[0][0]][tf.where(y1 == Y)[0][0]]
        q21 = inputs[tf.where(x2 == X)[0][0]][tf.where(y1 == Y)[0][0]]
        q12 = inputs[tf.where(x1 == X)[0][0]][tf.where(y2 == Y)[0][0]]
        q22 = inputs[tf.where(x2 == X)[0][0]][tf.where(y2 == Y)[0][0]]
        return (q11, q21, q12, q22), (x1, x2), (y1, y2)

    def _to_big(self, x_index, y_index):
        # TODO: Check index dtype
        # This must use different equations according to align_corners
        # set_trace()
        if self.align_corners:
            x_index = tf.linspace(0, len(x_index) * self.factor_upsample[0] - 1, len(x_index))
            y_index = tf.linspace(0, len(y_index) * self.factor_upsample[1] - 1, len(y_index))
        else:
            x_index = (x_index + 0.5) * self.factor_upsample[0] - 0.5
            y_index = (y_index + 0.5) * self.factor_upsample[1] - 0.5
            # x_index = tf.linspace(self.factor_upsample[0]/2 + 0.5, len(x_index)*self.factor_upsample[0] - algo,
            # len(x_index))
        return x_index, y_index  # Return the big index

    @staticmethod
    def _get_4_closest_points(x, y, x_list, y_list):
        # This function gets the 4 closests points
        x_distance = x_list - x
        y_distance = y_list - y
        if tf.math.reduce_any(x_distance >= 0):
            x_min = tf.math.reduce_min(tf.boolean_mask(x_distance, x_distance >= 0))
            x2 = tf.where(x_min == x_distance)[0][0]
        else:
            x2 = -1
        if tf.math.reduce_any(x_distance <= 0):
            x_min = tf.math.reduce_min(tf.math.abs(tf.boolean_mask(x_distance, x_distance <= 0)))
            x1 = tf.where(-x_min == x_distance)[0][0]
        else:
            x1 = -1
        if tf.math.reduce_any(y_distance >= 0):
            y_min = tf.math.reduce_min(tf.boolean_mask(y_distance, y_distance >= 0))
            y2 = tf.where(y_min == y_distance)[0][0]
        else:
            y2 = -1
        if tf.math.reduce_any(y_distance <= 0):
            y_min = tf.math.reduce_min(tf.math.abs(tf.boolean_mask(y_distance, y_distance <= 0)))
            y1 = tf.where(-y_min == y_distance)[0][0]
        else:
            y1 = -1
        if x2 == -1:
            x2 = x1
        if x1 == -1:
            x1 = x2
        if y2 == -1:
            y2 = y1
        if y1 == -1:
            y1 = y2
        return x_list[x2], x_list[x1], y_list[y2], y_list[y1]

    def get_real_equivalent(self):
        return ComplexUpSampling2D(size=self.factor_upsample, data_format=self.data_format,
                                   interpolation=self.interpolation, dtype=self.my_dtype.real_dtype)





# https://stackoverflow.com/questions/24100558/how-can-i-split-a-module-into-multiple-files-without-breaking-a-backwards-compa/24100645
from cvnn.layers.pooling import ComplexMaxPooling2D, ComplexAvgPooling2D
from cvnn.layers.pooling import ComplexUnPooling2D, ComplexMaxPooling2DWithArgmax
from cvnn.layers.convolutional import ComplexConv2D, ComplexConv1D, ComplexConv3D
from cvnn.layers.convolutional import ComplexConv2DTranspose
from cvnn.layers.core import ComplexInput, ComplexDense, ComplexFlatten, ComplexDropout, complex_input
from cvnn.layers.upsampling import ComplexUpSampling2D
from cvnn.layers.core import ComplexBatchNormalization


__author__ = 'J. Agustin BARRACHINA'
__copyright__ = 'Copyright 2020, {project_name}'
__credits__ = ['{credit_list}']
__license__ = '{license}'
__version__ = '1.0.12'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'
__status__ = '{dev_status}'



import logging
import pdb

import tensorflow as tf
from tensorflow.keras.layers import Activation
from typing import Union, Callable, Optional
from tensorflow import Tensor
from numpy import pi
# import cvnn

"""
This module contains many complex-valued activation functions to be used by CVNN class.
"""

# logger = logging.getLogger(cvnn.__name__)
t_activation = Union[str, Callable]  # TODO: define better


# Regression
def linear(z: Tensor) -> Tensor:
    """
    Does not apply any activation function. It just outputs the input.
    :param z: Input tensor variable
    :return: z
    """
    return z


def modrelu(z: Tensor, b: float = 1., c: float = 1e-3) -> Tensor:
    """
    mod ReLU presented in "Unitary Evolution Recurrent Neural Networks"
        from M. Arjovsky et al. (2016)
        URL: https://arxiv.org/abs/1511.06464
    A variation of the ReLU named modReLU. It is a pointwise nonlinearity,
    modReLU(z) : C -> C, which affects only the absolute
    value of a complex number, defined:
        modReLU(z) = ReLU(|z|+b)*z/|z|
    TODO: See how to check the non zero abs.
    """
    abs_z = tf.math.abs(z)
    return tf.cast(tf.keras.activations.relu(abs_z + b), dtype=z.dtype) * z / tf.cast(abs_z + c, dtype=z.dtype)


def zrelu(z: Tensor, epsilon=1e-7) -> Tensor:
    """
    zReLU presented in "On Complex Valued Convolutional Neural Networks"
        from Nitzan Guberman (2016).
    This methods let's the output as the input if both real and imaginary parts are positive.

    https://stackoverflow.com/questions/49412717/advanced-custom-activation-function-in-keras-tensorflow
    """
    imag_relu = tf.nn.relu(tf.math.imag(z))
    real_relu = tf.nn.relu(tf.math.real(z))
    ret_real = imag_relu*real_relu / (imag_relu + epsilon)
    ret_imag = imag_relu*real_relu / (real_relu + epsilon)
    ret_val = tf.complex(ret_real, ret_imag)
    return ret_val


def crelu(z: Tensor, alpha: float = 0.0, max_value: Optional[float] = None, threshold: float = 0) -> Tensor:
    """
    Mirror of cart_relu
    """
    return cart_relu(z, alpha, max_value, threshold)


def complex_cardioid(z: Tensor) -> Tensor:
    """
    Complex cardioid presented in "Better than Real: Complex-valued Neural Nets for MRI Fingerprinting"
        from V. Patrick (2017).
        
    This function maintains the phase information while attenuating the magnitude based on the phase itself. 
    For real-valued inputs, it reduces to the ReLU.
    """
    return tf.cast(1 + tf.math.cos(tf.math.angle(z)), dtype=z.dtype) * z / 2.

            
"""
Complex input, real output
"""


def cast_to_real(z: Tensor) -> Tensor:
    return tf.cast(z, z.dtype.real_dtype)


def sigmoid_real(z: Tensor) -> Tensor:
    return tf.keras.activations.sigmoid(tf.math.real(z) + tf.math.imag(z))


def softmax_real_with_abs(z: Tensor, axis=-1) -> Tensor:
    """
    Applies the softmax function to the modulus of z.
    The softmax activation function transforms the outputs so that all values are in range (0, 1) and sum to 1.
    It is often used as the activation for the last layer of a classification network because the result could be
    interpreted as a probability distribution.
    The softmax of x is calculated by exp(x)/tf.reduce_sum(exp(x)).
        https://www.tensorflow.org/api_docs/python/tf/keras/activations/softmax
    :param z: Input tensor.
    :return: Real-valued tensor of the applied activation function
    """
    if z.dtype.is_complex:
        return tf.keras.activations.softmax(tf.math.abs(z), axis)
    else:
        return tf.keras.activations.softmax(z, axis)


def softmax_real_with_avg(z: Tensor, axis=-1) -> Tensor:
    """
    The softmax activation function transforms the outputs so that all values are in range (0, 1) and sum to 1.
    It is often used as the activation for the last layer of a classification network because the result could be
    interpreted as a probability distribution.
    The softmax of x is calculated by exp(x)/tf.reduce_sum(exp(x)).
        https://www.tensorflow.org/api_docs/python/tf/keras/activations/softmax
    :param z: Input tensor.
    :return: Real-valued tensor of the applied activation function
    """
    if z.dtype.is_complex:
        return 0.5 * (tf.keras.activations.softmax(tf.math.real(z), axis) + tf.keras.activations.softmax(
            tf.math.real(z), axis))
    else:
        return tf.keras.activations.softmax(z, axis)


def softmax_real_with_mult(z: Tensor, axis=-1) -> Tensor:
    """
    Applies the softmax function to the modulus of z.
    The softmax activation function transforms the outputs so that all values are in range (0, 1) and sum to 1.
    It is often used as the activation for the last layer of a classification network because the result could be
    interpreted as a probability distribution.
    The softmax of x is calculated by exp(x)/tf.reduce_sum(exp(x)).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softmax
    :param z: Input tensor.
    :return: Real-valued tensor of the applied activation function
    """
    if z.dtype.is_complex:
        return tf.keras.activations.softmax(tf.math.real(z), axis) * tf.keras.activations.softmax(tf.math.real(z), axis)
    else:
        return tf.keras.activations.softmax(z, axis)


def softmax_of_softmax_real_with_mult(z: Tensor, axis=-1) -> Tensor:
    """
    Applies the softmax function to the modulus of z.
    The softmax activation function transforms the outputs so that all values are in range (0, 1) and sum to 1.
    It is often used as the activation for the last layer of a classification network because the result could be
    interpreted as a probability distribution.
    The softmax of x is calculated by exp(x)/tf.reduce_sum(exp(x)).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softmax
    :param z: Input tensor.
    :return: Real-valued tensor of the applied activation function
    """
    if z.dtype.is_complex:
        return tf.keras.activations.softmax(
            tf.keras.activations.softmax(tf.math.real(z), axis) * tf.keras.activations.softmax(tf.math.real(z), axis),
            axis)
    else:
        return tf.keras.activations.softmax(z, axis)


def softmax_of_softmax_real_with_avg(z: Tensor, axis=-1) -> Tensor:
    """
    Applies the softmax function to the modulus of z.
    The softmax activation function transforms the outputs so that all values are in range (0, 1) and sum to 1.
    It is often used as the activation for the last layer of a classification network because the result could be
    interpreted as a probability distribution.
    The softmax of x is calculated by exp(x)/tf.reduce_sum(exp(x)).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softmax
    :param z: Input tensor.
    :return: Real-valued tensor of the applied activation function
    """
    if z.dtype.is_complex:
        return tf.keras.activations.softmax(
            tf.keras.activations.softmax(tf.math.real(z), axis) + tf.keras.activations.softmax(tf.math.real(z), axis),
            axis)
    else:
        return tf.keras.activations.softmax(z, axis)


def softmax_real_by_parameter(z: Tensor, axis=-1, params: Optional[dict] = None) -> Tensor:
    if params is None:
        params = {
            'abs': True,
            'angle': True,
            'real': True,
            'imag': True
        }
    result = []
    for k, v in params:
        if k == 'abs' and v:
            result.append(tf.keras.activations.softmax(tf.math.abs(z), axis))
        if k == 'angle' and v:
            result.append(tf.keras.activations.softmax(tf.math.angle(z), axis))
        if k == 'real' and v:
            result.append(tf.keras.activations.softmax(tf.math.real(z), axis))
        if k == 'imag' and v:
            result.append(tf.keras.activations.softmax(tf.math.imag(z), axis))
    return tf.convert_to_tensor(result)


def convert_to_real_with_abs(z: Tensor) -> Tensor:
    """
    Applies the absolute value and returns a real-valued output.
    :param z: Input tensor.
    :return: Real-valued tensor of the applied activation function
    """
    if z.dtype.is_complex:
        return tf.math.abs(z)
    else:
        return z


def softmax_real_with_polar(z: Tensor, axis=-1) -> Tensor:
    """
    Applies the softmax function to the modulus of z.
    The softmax activation function transforms the outputs so that all values are in range (0, 1) and sum to 1.
    It is often used as the activation for the last layer of a classification network because the result could be
    interpreted as a probability distribution.
    The softmax of x is calculated by exp(x)/tf.reduce_sum(exp(x)).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softmax
    :param z: Input tensor.
    :return: Real-valued tensor of the applied activation function
    """
    if z.dtype.is_complex:
        return 0.5 * (tf.keras.activations.softmax(tf.math.abs(z), axis) + tf.keras.activations.softmax(tf.math.angle(z),
                                                                                                        axis))
    else:
        return tf.keras.activations.softmax(z, axis)


"""
etf Functions
"""


def etf_circular_tan(z: Tensor) -> Tensor:
    return tf.math.tan(z)


def etf_circular_sin(z: Tensor) -> Tensor:
    return tf.math.sin(z)


def etf_inv_circular_atan(z: Tensor) -> Tensor:
    return tf.math.atan(z)


def etf_inv_circular_asin(z: Tensor) -> Tensor:
    return tf.math.asin(z)


def etf_inv_circular_acos(z: Tensor) -> Tensor:
    return tf.math.acos(z)


def etf_circular_tanh(z: Tensor) -> Tensor:
    return tf.math.tanh(z)


def etf_circular_sinh(z: Tensor) -> Tensor:
    return tf.math.sinh(z)


def etf_inv_circular_atanh(z: Tensor) -> Tensor:
    return tf.math.atanh(z)


def etf_inv_circular_asinh(z: Tensor) -> Tensor:
    return tf.math.asinh(z)


"""
Phasor Networks
"""


def georgiou_cdbp(z:Tensor, r: float = 1, c: float = 1e-3) -> Tensor:
    """
    Activation function proposed by G. M. Georgioy and C. Koutsougeras in
        https://ieeexplore.ieee.org/abstract/document/142037
    """
    return z / tf.cast(c + tf.math.abs(z)/r, dtype=z.dtype)


def complex_signum(z: Tensor, k: Optional[int] = None) -> Tensor:
    """
    Complex signum activation function is very similar to mvn_activation.
    For a detailed explanation refer to:
        https://ieeexplore.ieee.org/abstract/document/548176
    """
    if k:
        # values = np.linspace(pi / k, 2 * pi - pi / k, k)
        angle_cast = tf.math.floor(tf.math.angle(z) * k / (2 * pi))
        # import pdb; pdb.set_trace()
        return tf.math.exp(tf.complex(
            tf.zeros(tf.shape(z), dtype=z.dtype.real_dtype), angle_cast * 2 * pi / k))
    else:
        return tf.math.exp(tf.complex(tf.zeros(tf.shape(z), dtype=z.dtype.real_dtype), tf.math.angle(z)))


def mvn_activation(z: Tensor, k: Optional[int] = None) -> Tensor:
    """
    Function inspired by Naum Aizenberg.
        A multi-valued neuron (MVN) is a neural element with n inputs and one output lying on the unit circle,
        and with complex-valued weights.
    Works:
        https://link.springer.com/article/10.1007%2FBF01068667
        http://pefmath2.etf.rs/files/93/399.pdf
    """
    if k:
        # values = np.linspace(pi / k, 2 * pi - pi / k, k)
        angle_cast = tf.math.floor(tf.math.angle(z) * k / (2 * pi))
        # import pdb; pdb.set_trace()
        return tf.math.exp(tf.complex(
            tf.zeros(tf.shape(z), dtype=z.dtype.real_dtype), (angle_cast + 0.5) * 2 * pi / k))
    else:
        return tf.math.exp(tf.complex(tf.zeros(tf.shape(z), dtype=z.dtype.real_dtype), tf.math.angle(z)))


"""
TYPE A: Cartesian form.
"""


# TODO: shall I use tf.nn or tf.keras.activation modules?
# https://stackoverflow.com/questions/54761088/tf-nn-relu-vs-tf-keras-activations-relu
# nn has leaky relu, activation doesn't


def cart_sigmoid(z: Tensor) -> Tensor:
    """
    Applies the function (1.0 / (1.0 + exp(-x))) + j * (1.0 / (1.0 + exp(-y))) where z = x + j * y
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/sigmoid
    :param z: Tensor to be used as input of the activation function
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.sigmoid(tf.math.real(z)),
                              tf.keras.activations.sigmoid(tf.math.imag(z))), 
                   dtype=z.dtype)


def cart_elu(z: Tensor, alpha=1.0) -> Tensor:
    """
    Applies the "Exponential linear unit": x if x > 0 and alpha * (exp(x)-1) if x < 0
    To both the real and imaginary part of z.
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/elu
    :param z: Input tensor.
    :param alpha: A scalar, slope of negative section.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.elu(tf.math.real(z), alpha),
                              tf.keras.activations.elu(tf.math.imag(z), alpha)), dtype=z.dtype)


def cart_exponential(z: Tensor) -> Tensor:
    """
    Exponential activation function. Applies to both the real and imag part of z the exponential activation: exp(x)
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/exponential
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.exponential(tf.math.real(z)),
                              tf.keras.activations.exponential(tf.math.imag(z))), dtype=z.dtype)


def cart_hard_sigmoid(z: Tensor) -> Tensor:
    """
    Applies the Hard Sigmoid function to both the real and imag part of z.
    The hard sigmoid function is faster to compute than sigmoid activation.
    Hard sigmoid activation:    0               if x < -2.5
                                1               if x > 2.5
                                0.2 * x + 0.5   if -2.5 <= x <= 2.5
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/hard_sigmoid
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.hard_sigmoid(tf.math.real(z)),
                              tf.keras.activations.hard_sigmoid(tf.math.imag(z))), dtype=z.dtype)


def cart_relu(z: Tensor, alpha: float = 0.0, max_value: Optional[float] = None, threshold: float = 0) -> Tensor:
    """
    Applies Rectified Linear Unit to both the real and imag part of z
    The relu function, with default values, it returns element-wise max(x, 0).
    Otherwise, it follows:  f(x) = max_value for x >= max_value,
                            f(x) = x for threshold <= x < max_value,
                            f(x) = alpha * (x - threshold) otherwise.
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/relu
    :param z: Tensor -- Input tensor.
    :param alpha: float -- A float that governs the slope for values lower than the threshold (default 0.0).
    :param max_value: Optional float -- A float that sets the saturation threshold (the largest value the function will return)
        (default None).
    :param threshold: float -- A float giving the threshold value of the activation function below which
        values will be damped or set to zero (default 0).
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.relu(tf.math.real(z), alpha, max_value, threshold),
                              tf.keras.activations.relu(tf.math.imag(z), alpha, max_value, threshold)), dtype=z.dtype)


def cart_leaky_relu(z: Tensor, alpha=0.2, name=None) -> Tensor:
    """
    Applies Leaky Rectified Linear Unit to both the real and imag part of z
    https://www.tensorflow.org/api_docs/python/tf/nn/leaky_relu
    http://robotics.stanford.edu/~amaas/papers/relu_hybrid_icml2013_final.pdf
    :param z: Input tensor.
    :param alpha: Slope of the activation function at x < 0. Default: 0.2
    :param name: A name for the operation (optional).
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.nn.leaky_relu(tf.math.real(z), alpha, name),
                              tf.nn.leaky_relu(tf.math.imag(z), alpha, name)), dtype=z.dtype)


def cart_selu(z: Tensor) -> Tensor:
    """
    Applies Scaled Exponential Linear Unit (SELU) to both the real and imag part of z.
    The scaled exponential unit activation: scale * elu(x, alpha).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/selu
    https://arxiv.org/abs/1706.02515
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.selu(tf.math.real(z)),
                              tf.keras.activations.selu(tf.math.imag(z))), dtype=z.dtype)


def cart_softplus(z: Tensor) -> Tensor:
    """
    Applies Softplus activation function to both the real and imag part of z.
    The Softplus function: log(exp(x) + 1)
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softplus
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.softplus(tf.math.real(z)),
                              tf.keras.activations.softplus(tf.math.imag(z))), dtype=z.dtype)


def cart_softsign(z: Tensor) -> Tensor:
    """
    Applies Softsign activation function to both the real and imag part of z.
    The softsign activation: x / (abs(x) + 1).      TODO: typo in tensorflow references (softplus instead of softsign)
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softsign
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.softsign(tf.math.real(z)),
                              tf.keras.activations.softsign(tf.math.imag(z))), dtype=z.dtype)


def cart_tanh(z: Tensor) -> Tensor:
    """
    Applies Hyperbolic Tangent (tanh) activation function to both the real and imag part of z.
    The tanh activation: tanh(x) = sinh(x)/cosh(x) = ((exp(x) - exp(-x))/(exp(x) + exp(-x))).
    The derivative if tanh is computed as 1 - tanh^2 so it should be fast to compute for backprop.
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/tanh
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.tanh(tf.math.real(z)),
                              tf.keras.activations.tanh(tf.math.imag(z))), dtype=z.dtype)


# Classification
def cart_softmax(z: Tensor, axis=-1) -> Tensor:
    """
    Applies the softmax function to both the real and imag part of z.
    The softmax activation function transforms the outputs so that all values are in range (0, 1) and sum to 1.
    It is often used as the activation for the last layer of a classification network because the result could be
    interpreted as a probability distribution.
    The softmax of x is calculated by exp(x)/tf.reduce_sum(exp(x)).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/softmax
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return tf.cast(tf.complex(tf.keras.activations.softmax(tf.math.real(z), axis),
                              tf.keras.activations.softmax(tf.math.imag(z), axis)), dtype=z.dtype)


"""
TYPE B: Polar form.
"""


# For all ReLU functions, the polar form makes no real sense. If we keep the phase because abs(z) > 0


def _apply_pol(z: Tensor, amp_fun: Callable[[Tensor], Tensor],
               pha_fun: Optional[Callable[[Tensor], Tensor]] = None) -> Tensor:
    amp = amp_fun(tf.math.abs(z))
    pha = tf.math.angle(z)
    if pha_fun is not None:
        pha = pha_fun(pha)
    return tf.cast(tf.complex(amp * tf.math.cos(pha), amp * tf.math.sin(pha)), dtype=z.dtype)


def pol_tanh(z: Tensor) -> Tensor:
    """
    Applies Hyperbolic Tangent (tanh) activation function to the amplitude of the complex number
        leaving the phase untouched.
    The derivative if tanh is computed as 1 - tanh^2 so it should be fast to compute for backprop.
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/tanh
    :param z: Input tensor.
    :return: Tensor result of the applied activation function
    """
    return _apply_pol(z, tf.keras.activations.tanh)


def pol_sigmoid(z: Tensor) -> Tensor:
    """
    Applies the sigmoid function to the amplitude of the complex number leaving the phase untouched
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/sigmoid
    :param z: Tensor to be used as input of the activation function
    :return: Tensor result of the applied activation function
    """
    return _apply_pol(z, tf.keras.activations.sigmoid)


def pol_selu(z: Tensor) -> Tensor:
    """
    Applies Scaled Exponential Linear Unit (SELU) to the absolute value of z, keeping the phase unchanged.
    The scaled exponential unit activation: scale * elu(x, alpha).
    https://www.tensorflow.org/api_docs/python/tf/keras/activations/selu
    https://arxiv.org/abs/1706.02515
    :param z: Input tensor.
    :return: Tensor result of the applied activation function

    Logic:
        I must mantain the phase (angle) so: cos(theta) = x_0/r_0 = x_1/r_1.
        For real case, x_0 = r_0 so it also works.
    """
    r_0 = tf.abs(z)
    r_1 = tf.keras.activations.selu(r_0)
    return tf.cast(tf.complex(tf.math.real(z) * r_1 / r_0, tf.math.imag(z) * r_1 / r_0), dtype=z.dtype)


act_dispatcher = {
    'linear': Activation(linear),
    # Complex input, real output
    'cast_to_real': Activation(cast_to_real),
    'convert_to_real_with_abs': Activation(convert_to_real_with_abs),
    'sigmoid_real': Activation(sigmoid_real),
    'softmax_real_with_abs': Activation(softmax_real_with_abs),
    'softmax_real_with_avg': Activation(softmax_real_with_avg),
    'softmax_real_with_mult': Activation(softmax_real_with_mult),
    'softmax_of_softmax_real_with_mult': Activation(softmax_of_softmax_real_with_mult),
    'softmax_of_softmax_real_with_avg': Activation(softmax_of_softmax_real_with_avg),
    'softmax_real_with_polar': Activation(softmax_real_with_polar),
    # Phasor networks
    'georgiou_cdbp': Activation(georgiou_cdbp),
    'mvn_activation': Activation(mvn_activation),
    'complex_signum': Activation(complex_signum),
    # Type A (cartesian)
    'cart_sigmoid': Activation(cart_sigmoid),
    'cart_elu': Activation(cart_elu),
    'cart_exponential': Activation(cart_exponential),
    'cart_hard_sigmoid': Activation(cart_hard_sigmoid),
    'cart_relu': Activation(cart_relu),
    'cart_leaky_relu': Activation(cart_leaky_relu),
    'cart_selu': Activation(cart_selu),
    'cart_softplus': Activation(cart_softplus),
    'cart_softsign': Activation(cart_softsign),
    'cart_tanh': Activation(cart_tanh),
    'cart_softmax': Activation(cart_softmax),
    # Type B (polar)
    'pol_tanh': Activation(pol_tanh),
    'pol_sigmoid': Activation(pol_sigmoid),
    'pol_selu': Activation(pol_selu),
    # Elementary Transcendental Functions (ETF)
    'etf_circular_tan': Activation(etf_circular_tan),
    'etf_circular_sin': Activation(etf_circular_sin),
    'etf_inv_circular_atan': Activation(etf_inv_circular_atan),
    'etf_inv_circular_asin': Activation(etf_inv_circular_asin),
    'etf_inv_circular_acos': Activation(etf_inv_circular_acos),
    'etf_circular_tanh': Activation(etf_circular_tanh),
    'etf_circular_sinh': Activation(etf_circular_sinh),
    'etf_inv_circular_atanh': Activation(etf_inv_circular_atanh),
    'etf_inv_circular_asinh': Activation(etf_inv_circular_asinh),
    # ReLU
    'modrelu': Activation(modrelu),
    'crelu': Activation(crelu),
    'zrelu': Activation(zrelu),
    'complex_cardioid': Activation(complex_cardioid)
}

if __name__ == '__main__':
    x = tf.constant([-2, 1.0, 0.0, 1.0, -3, 0.8, 0.1], dtype=tf.float32)
    y = tf.constant([-2.5, -1.5, 0.0, 1.0, 2, 0.4, -0.4], dtype=tf.float32)
    z = tf.complex(x, y)
    result = crelu(z)
    result = modrelu(z, 4)
    result = zrelu(z)
    result = complex_cardioid(z)
    """import matplotlib.pyplot as plt
    import numpy as np

    x = tf.constant([-2, 1.0, 0.0, 1.0, -3, 0.8, 0.1], dtype=tf.float32)
    y = tf.constant([-2.5, -1.5, 0.0, 1.0, 2, 0.4, -0.4], dtype=tf.float32)
    z = tf.complex(x, y)
    result = georgiou_cdbp(z)

    ax = plt.axes()
    ax.scatter(tf.math.real(z), tf.math.imag(z), color='red')
    ax.scatter(tf.math.real(result), tf.math.imag(result), color='blue')
    for x, y, dx, dy in zip(tf.math.real(z), tf.math.imag(z),
                            tf.math.real(result) - tf.math.real(z),
                            tf.math.imag(result) - tf.math.imag(z)):
        ax.arrow(x, y, dx, dy, length_includes_head=True, head_width=0.1)
    t = np.linspace(0, np.pi * 2, 100)
    ax.plot(np.cos(t), np.sin(t), linewidth=1)

    yabs_max = abs(max(ax.get_ylim(), key=abs))
    xabs_max = abs(max(ax.get_xlim(), key=abs))
    axis_max = max(yabs_max, xabs_max)

    ax.set_ylim(ymin=-axis_max, ymax=axis_max)
    ax.set_xlim(xmin=-axis_max, xmax=axis_max)
    plt.show()"""
    

__author__ = 'J. Agustin BARRACHINA'
__version__ = '0.0.19'
__maintainer__ = 'J. Agustin BARRACHINA'
__email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'



from setuptools import setup
import versioneer

requirements = [
    'tensorflow>=2.0', 'tensorflow-probability',    # tfp for the Batch Norm (covariance)
    'numpy', 'six', 'packaging',
    'pandas', 'scipy',                   # Data
    'colorlog', 'openpyxl',              # Logging
    'tqdm'
]

setup(
    name='cvnn',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Library to help implement a complex-valued neural network (cvnn) using tensorflow as back-end",
    license="MIT",
    author="J Agustin BARRACHINA",
    author_email='joseagustin.barra@gmail.com',
    url='https://github.com/NEGU93/cvnn',
    packages=['cvnn', 'cvnn.layers'],
    entry_points={
        'console_scripts': [
            'cvnn=cvnn.cli:cli'
        ]
    },
    install_requires=requirements,
    keywords='cvnn',
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    long_description_content_type="text/markdown",
    long_description=open('README.md').read(),
    extras_require={
        'plotter': ['matplotlib', 'seaborn', 'plotly', 'tikzplotlib'],
        'full': ['prettytable', 'matplotlib', 'seaborn', 'plotly', 'tikzplotlib']
    }

)



from tensorflow.keras.callbacks import TensorBoard
from tensorflow import GradientTape
import tensorflow as tf


# This extends TensorBoard to save gradients as histogram
# ExtendedTensorBoard can then be used in replace of tf.keras.callbacks.TensorBoard.
class ExtendedTensorBoard(TensorBoard):
    def _log_gradients(self, epoch):
        writer = self._writers['train']

        with writer.as_default(), GradientTape() as g:
            # here we use test data to calculate the gradients
            features, y_true = list(dataset.batch(100).take(1))[0]

            y_pred = self.model(features)  # forward-propagation
            loss = self.model.compiled_loss(y_true=y_true, y_pred=y_pred)  # calculate loss
            gradients = g.gradient(loss, self.model.trainable_weights)  # back-propagation

            # In eager mode, grads does not have name, so we get names from model.trainable_weights
            for weights, grads in zip(self.model.trainable_weights, gradients):
                tf.summary.histogram(
                    weights.name.replace(':', '_') + '_grads', data=grads, step=epoch)

        writer.flush()

    def on_epoch_end(self, epoch, logs=None):
        # This function overwrites the on_epoch_end in tf.keras.callbacks.TensorBoard
        # but we do need to run the original on_epoch_end, so here we use the super function.
        super(ExtendedTensorBoard, self).on_epoch_end(epoch, logs=logs)

        if self.histogram_freq and epoch % self.histogram_freq == 0:
            self._log_gradients(epoch)
